program HtmlTableToDosText;

{
****************************************************************
* Name:			    HtmlTableToDosText                 *
* Platform:		    Delphi 7.0			       *
* Version:                  0.1                                *
* Author:                   k561ln7                            *
* E-Mail:                   k561ln7@pochta.ru                  *
* Date of Create:           28.03.2006                         *
* Last Updating:            28.03.2006                         *
****************************************************************
}

{$APPTYPE CONSOLE}

uses
  Windows, SysUtils, Classes;

type
  TFileCod = (fcAnsi, fcUTF8, fcOem);

const
  arDelTag     :array[0..10] of String = ('<B>','</B>','&NBSP;','<I>','</I>',#13,#10,'</FONT>','<U>','</U>','</A>');
  arDelTagLen  :array[0..10] of Word   = (3    ,4     ,6       ,3    ,4     ,1  ,1  ,7        ,3    ,4     ,4);
  arDelTagN    :array[0..1] of String  = ('<FONT','<A');
  arDelTagLenN :array[0..1] of Word    = (5     ,2);

  shelp = #13#10+
  '       HtmlTableToDosText 0.1 '#13#10+
  ' (c)k561ln7 E-Mail:k561ln7@pochta.ru'#13#10+
  '                          28.03.2006'#13#10#13#10+
  ' Usage: HtmlTableToDosText <file> [options]'#13#10
  + ' Options:'#13#10
  + '   /code:[ansi, utf8, oem] - default is "oem"'#13#10
  + '   /maxcolsize:[0..255]    - default is "0"'#13#10
  + '   /horline:[+,-]          - default is "-"';

var
  arTR        :array of String = nil;
  arTD        :array of array of String = nil;
  FileCod     :TFileCod = fcOem;
  iTable      :Integer = 0;
  iMaxColSize :Byte    = 0;
  bNoHorLine  :boolean = false;

function UTF82Ansi(const Source: string): WideString;
var
  Iterator,
  SourceLength,
  FChar, NChar,
  ResLen,
  MemLen       :Integer;

  procedure AddSimbolInRes(C:WideChar);
  //use
  //  ResLen :Integer;
  //  MemLen :Integer;
  //  result :WideString;
  begin
    inc(ResLen);
    if (ResLen mod MemLen) = 0 then
      SetLength(result, ResLen+MemLen);
    Result[ResLen] := C;
  end;

  function CharToWideChar(C: Char): WideChar;
  var
    S: WideString;
  begin
    S := C;
    Result := S[1];
  end;

begin
  Result := '';
  Iterator := 0;
  ResLen := 0;
  SourceLength := Length(Source);
  MemLen := SourceLength div 10;
  SetLength(result, ResLen+MemLen);
  while Iterator < SourceLength do
  begin
    Inc(Iterator);
    FChar := Ord(Source[Iterator]);
    if FChar >= $80 then
    begin
      Inc(Iterator);
      if Iterator > SourceLength then break;
      FChar := FChar and $3F;
      if (FChar and $20) <> 0 then
      begin
        FChar := FChar and $1F;
        NChar := Ord(Source[Iterator]);
        if (NChar and $C0) <> $80 then
        begin
          AddSimbolInRes(CharToWideChar(Source[Iterator-1]));
          AddSimbolInRes(CharToWideChar(Source[Iterator]));
          continue;
        end;
        FChar := (FChar shl 6) or (NChar and $3F);
        Inc(Iterator);
        if Iterator > SourceLength then break;
      end;
      NChar := Ord(Source[Iterator]);
      if (NChar and $C0) <> $80 then
      begin
        AddSimbolInRes(CharToWideChar(Source[Iterator-1]));
        AddSimbolInRes(CharToWideChar(Source[Iterator]));
      end
      else
        AddSimbolInRes(WideChar((FChar shl 6) or (NChar and $3F)));
    end
    else
      AddSimbolInRes(WideChar(FChar));
  end;
  SetLength(result, ResLen);
end;

function Ansi2Oem(const Ansi: String): String;
var
 outBuff: PChar;
begin
  result := Ansi;
  if Length(Ansi) = 0 then exit;
  outBuff:=StrAlloc(Length(Ansi)+1);
  CharToOem(PChar(Ansi), outBuff);
  Result:=StrPas(outBuff);
  StrDispose(outBuff);
end;

{
function Oem2Ansi(const Oem: String): String;
var
 outBuff: PChar;
begin
  result := Oem;
  if Length(Oem) = 0 then exit;
  outBuff:=StrAlloc(Length(Oem)+1);
  OemToChar(PChar(Oem), outBuff);
  Result:=StrPas(outBuff);
  StrDispose(outBuff);
end;
}

function DelTag(Str: String): String;
//use
// arDelTag: array[0..6] of String
// arDelTagLen: array[0..6] of Word
// arDelTagN: array[0..1] of String
// arDelTagLenN: array[0..1] of Word
var
 i, j,
 iLenStr,
 iPosStr,
 iPosRes :Integer;
 b       :Boolean;
begin
  iLenStr := Length(Str);
  SetLength(result, iLenStr);
  iPosRes := 0;
  iPosStr := 1;
  repeat
    b := false;
    for i := 0 to High(arDelTag) do
      if arDelTag[i] = UpperCase(copy(Str, iPosStr, arDelTagLen[i])) then
      begin
        inc(iPosStr, arDelTagLen[i]);
        b := true;
        break;
      end;
    if b then continue;
    for i := 0 to High(arDelTagN) do
      if arDelTagN[i] = UpperCase(copy(Str, iPosStr, arDelTagLenN[i])) then
      begin
        for j := iPosStr+arDelTagLen[i] to iLenStr do
          if Str[j] = '>' then
            break;
        iPosStr := j+1;
        b := true;
        break;
      end;
    if b then continue;
    inc(iPosRes);
    result[iPosRes] := Str[iPosStr];
    inc(iPosStr);
  until iPosStr > iLenStr;
  SetLength(result, iPosRes);
end;

function TagCut(Const Str, BeginTag, EndTag: String; var PosStr: Integer; var StrOut: String): Boolean;
var
 j,
 iLenBeginTag,
 iLenEndTag,
 iLenStr      :Integer;
 b            :Boolean;
begin
  iLenStr := Length(Str);
  iLenBeginTag := Length(BeginTag);
  iLenEndTag := Length(EndTag);
  b := false;
  result := false;
  StrOut := '';
  repeat
    if BeginTag = UpperCase(copy(Str, PosStr, iLenBeginTag)) then
    begin
      b := true;
      for j := PosStr+iLenBeginTag to iLenStr do
        if Str[j] = '>' then
          break;
      PosStr := j+1;
    end;
    if EndTag = UpperCase(copy(Str, PosStr, iLenEndTag)) then
    begin
      result := true;
      inc(PosStr, iLenEndTag);
      break;
    end;

    if b then
      StrOut := StrOut+Str[PosStr];
    inc(PosStr);
  until PosStr >= iLenStr;
end;

procedure ArrayTR(const Str: String);
//use
//  arTR: array of String;
//  iTable: Integer;
var
 i,
 iarTRLen,
 iPerc,
 iLenStr  :Integer;
 S        :String;
begin
  i := 1;
  S := '';
  iarTRLen := 0;
  iLenStr := Length(Str);
  iPerc := 0;
  SetConsoleTitle(PChar(Format('Table %d: Create TR array %d%%',[iTable, iPerc])));
  repeat
    if not TagCut(Str, '<TR','</TR>',i,S) then
      continue;
    if (iarTRLen mod 80) = 0 then
      SetLength(arTR, iarTRLen+80);
    arTR[iarTRLen] := S;
    inc(iarTRLen);
    if iPerc < Round(i / iLenStr * 100) then
    begin
      iPerc := Round(i / iLenStr * 100);
      SetConsoleTitle(PChar(Format('Table %d: Create TR array %d%%',[iTable, iPerc])));
    end;
  until i >= iLenStr;
  SetLength(arTR, iarTRLen);
end;

procedure ArrayTD;
//use
//  arTR: array of String;
//  arTD: array of String;
//  iTable: Integer;
//  iMaxColSize: Integer = 0;
var
 i,j,
 iarTRLen,
 iarTDLen,
 iPerc    :Integer;
 S        :String;
begin
  S := '';
  SetLength(arTD, Length(arTR));
  iarTRLen := High(arTR);
  iPerc := 0;
  SetConsoleTitle(PChar(Format('Table %d: Create TD array %d%%',[iTable, iPerc])));
  for j := 0 to iarTRLen do
  begin
    i := 1;
    iarTDLen := 0;
    repeat
      if not TagCut(arTR[j], '<TD','</TD>',i,S) then
        continue;
      if (iarTDLen mod 30) = 0 then
        SetLength(arTD[j], iarTDLen+30);
      S := DelTag(S);
      if (iMaxColSize > 0)and(Length(S)>iMaxColSize) then
      begin
        arTD[j][iarTDLen] := copy(S,1,iMaxColSize);
        arTD[j][iarTDLen][iMaxColSize] := #16;
      end
      else
        arTD[j][iarTDLen] := S;
      inc(iarTDLen);
    until i >= Length(arTR[j]);
    SetLength(arTD[j], iarTDLen);
    if iPerc < Round(j / iarTRLen * 100) then
    begin
      iPerc := Round(j / iarTRLen * 100);
      SetConsoleTitle(PChar(Format('Table %d: Create TD array %d%%',[iTable, iPerc])));
    end;
  end;
end;

function AddSpice(const Str: String; StrLen: Integer): String;
var
  i :Integer;
  S :String;
begin
  i := Length(Str);
  if StrLen < i then
  begin
    Result := Str;
    Exit;
  end;
  SetLength(S, StrLen-i);
  FillChar(S[1], Length(S), ' ');
  Result := Str+S;
end;

var
 i, j, iLen,
 iMaxLenarTD,
 itabBegin, itabEnd,
 iarTDLen           :Integer;
 sNameFile,
 sLine1, sLine2,
 S, Str             :String;
 arSizeCol          :array of Word;
 FileSource         :TFileStream;
begin
  if ParamCount = 0 then
  begin
    WriteLn(shelp);
    exit;
  end;

  i := 0;
  repeat
    inc(i);
    S := UpperCase(ParamStr(i));
    if Trim(S) = '' then continue;

    if (S = '/CODE:ANSI') then
    begin
      FileCod := fcAnsi;
      continue;
    end;

    if (S = '/CODE:UTF8') then
    begin
      FileCod := fcUTF8;
      continue;
    end;

    if (S = '/CODE:OEM') then
    begin
      FileCod := fcOEM;
      continue;
    end;

    if Pos('/MAXCOLSIZE:',S) > 0 then
    begin
      if (Length(S) <= Length('/MAXCOLSIZE:')) then continue;
      j := Pos('/MAXCOLSIZE:',S)+Length('/MAXCOLSIZE:');
      iMaxColSize := StrToIntDef(copy(S,j,Length(S)-j+1),0);
      continue;
    end;

    if Pos('/HORLINE:',S) > 0 then
    begin
      if (Length(S) > Length('/HORLINE:'))
         and (S[Length('/HORLINE:')+1] = '-')
      then
        bNoHorLine := true;
      continue;
    end;


    if FileExists(S) then
    begin
      sNameFile := S;
      continue;
    end;

    WriteLn('Error: Invalid option: '+S);
    Exit;

  until i > ParamCount;

  FileSource := TFileStream.Create(sNameFile, fmOpenRead);
  SetLength(S, FileSource.Size);
  FileSource.Read(S[1],FileSource.Size);
  FileSource.Free;

  if Trim(S) = '' then Exit;

  SetConsoleTitle('File Decode');

  SetLength(Str, Length(S));
  j := 1;
  for i := 1 to Length(S) do //   , Ansi2Oem 
    if S[i] <> #0 then       //     
    begin                    //xlhtml   xls    
      Str[j] := S[i];
      inc(j);
    end;
  SetLength(Str, j);
  S := '';

  case FileCod of
    fcAnsi: Str := Ansi2Oem(Str);
    fcUTF8: Str := Ansi2Oem(UTF82Ansi(Str));
  end;

  iTable := 0;
  repeat
    S := UpperCase(Str);
    itabBegin := pos('<TABLE', S);
    itabEnd := pos('</TABLE>', S);
    S := '';

    if (itabBegin <= 0) or (itabEnd <= 0) then continue;


    inc(iTable);
    SetConsoleTitle(PChar(Format('Table %d: Create TR array',[iTable])));
    ArrayTR(copy(Str,itabBegin,itabEnd+8-itabBegin));

    SetConsoleTitle(PChar(Format('Table %d: Create TD array',[iTable])));
    ArrayTD;

    Delete(Str,itabBegin,itabEnd+8-itabBegin);

    arTR := nil;

    if (Length(arTD) <= 0) then
    begin
      arTD := nil;
      Continue;
    end;

    SetConsoleTitle(PChar(Format('Table %d: Create Size Columns array',[iTable])));

    iarTDLen := High(arTD);

    iMaxLenarTD := 0;
    for i := 0 to iarTDLen do
      if Length(arTD[i]) > iMaxLenarTD then
        iMaxLenarTD := i;

    if (Length(arTD[iMaxLenarTD]) <= 0) then
    begin
      arTD := nil;
      Continue;
    end;

    SetLength(arSizeCol, Length(arTD[iMaxLenarTD]));
    for i := 0 to High(arTD[iMaxLenarTD]) do
    begin
      iLen := 0;
      for j := 0 to iarTDLen do
      begin
        if High(arTD[j]) < i then continue;
        S := arTD[j][i];
        if Length(S) > iLen then
          iLen := Length(S);
      end;
      arSizeCol[i] := iLen;
    end;

    SetConsoleTitle(PChar(Format('Table %d: Write',[iTable])));
    for i := 0 to iarTDLen do
    begin
      S := '';
      sLine2 := '';
      for j:=0 to High(arTD[i]) do
      begin
        sLine1 := AddSpice(arTD[i][j],arSizeCol[j])+#179;
        S := S+sLine1;
        if not bNoHorLine then
        begin
          FillChar(sLine1[1],Length(sLine1)-1,#196);
          sLine1[Length(sLine1)] := #197;
          sLine2 := sLine2+sLine1;
        end;
      end;
      WriteLn(S);
      if not bNoHorLine then
      begin
        WriteLn(sLine2);
        if (i mod 5) = 0 then
          SetConsoleTitle(PChar(Format('Table %d: Write Line: %d/%d',[iTable,i*2,iarTDLen*2])));
      end else
        if (i mod 10) = 0 then
          SetConsoleTitle(PChar(Format('Table %d: Write Line: %d/%d',[iTable,i,iarTDLen])));
    end;

    WriteLn(#13#10);

    arTD := nil;
  until (itabBegin <= 0) or (itabEnd <= 0);
end.
